library vasat;

import 'dart:convert';
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart' as http;
import 'package:shared_preferences/shared_preferences.dart';
import 'dart:async';
import 'models/vasat_model.dart';
import 'vasat_config.dart';
import 'models/session.dart';

/// Vasat flutter SDK
class Vasat {

  String name;
  VasatConfig config;
  Session _session;

  Session get currentSession => _session;

  static const String defaultName = "vasat";
  // _cache is library-private, thanks to the _ in front of its name.
  static final Map<String, Vasat> _cache = <String, Vasat>{};


  factory Vasat({String name = defaultName, VasatConfig config}) {
    // Return cached vasat for given name. if not present calls Vasat._init to generate one.
    return _cache.putIfAbsent(
        name, () => Vasat._init(name, config));
  }
  /// Additional method that does the same as the factory constructor. Just here for making it more natural to get instances of Vasat
  static Vasat instance([String name = defaultName]) {
    // Return cached vasat for given name. if not present calls Vasat._init to generate one.
    return _cache[name];
  }

  Vasat._init(this.name, this.config) {
    // initialization logic here
  }

  Future<bool> hasActiveSession() {
    return getActiveSession().then((session) {
      if (session != null && session.expiry == 0) {
        _session = session;
        return true;
      } else {
        _session = null;
        return false;
      }
    });
  }

  Future<Session> getActiveSession() {
    return getObjects<Session>(Session()).then((values) {
      if (values != null && values.length > 0) {
        return values[0];
      } else {
        return null;
      }
    });
  }

  Future<Session> loginWithClientCredentials() async {
    var body = jsonEncode(<String, String>{
      "grant_type": "client_credentials",
      "client_id": this.config?.clientId,
      "client_secret": this.config?.clientSecret,
    });
    debugPrint("ENTER body to send: " + body);
    return apiPost("/access/token", body).then((ses) {
      // Store current session
      _session = ses as Session;
      return _session;
    });
  }

  Future<dynamic> apiPost(String url, dynamic params) async {
    if (this.config != null && this.config.host != null) {
      return http.post(this.config.host + url, body: params, headers: getApiHeaders(isJson: true))
          .then((res) => processResponse(res));
    } else  {
      return Future.error("No vasat configuration or host found");
    }

  }

  Future<dynamic> apiGet(String url, [Map<String, String> params]) async {
    if (this.config != null && this.config.host != null) {
      var uri = Uri.parse(this.config.host + url);
      if (params != null) uri = uri.replace(queryParameters: params);

      debugPrint("ENTER apiGet URL: ${uri.toString()} With params: ${uri.queryParameters.toString()}");

      return http.get(uri, headers: getApiHeaders())
          .then((res) => processResponse(res));
    } else  {
      return Future.error("No vasat configuration or host found");
    }

  }

  Map<String, String> getApiHeaders({bool isJson = false}) {
    var headers = <String,String>{};
    // If POST or PUT methods this should be set to true
    if (isJson) headers["Content-Type"] = "application/json";
    // Check if we have token
    debugPrint("ENTER getApiHeaders checking session: " + _session.toString());

    if (_session?.token != null) headers["Authorization"] = "Bearer " + _session.token;

    return headers;
  }

  Future<dynamic> processResponse(http.Response response) async {

    // Always expect JSON payload. If not JSON throw an error.
    try {
      Map<String, dynamic> resBody = jsonDecode(response.body);
      //debugPrint("ENTER processResponse resBody:" + resBody.toString());

      if (response.statusCode < 400 && resBody.containsKey("access_token") ) {
        var ses = Session().fromJson(resBody);
        await saveObjects([ses]);
        // Process access token
        return Future<dynamic>.value( ses );
      } else if (response.statusCode < 400 && resBody.containsKey("payload") && resBody.containsKey("status") && resBody["status"] < 400 ) {
        return Future<dynamic>.value(resBody["payload"]);
      } else if (response.statusCode < 400 && resBody.containsKey("status") && resBody["status"] < 400 ) {
        return Future<dynamic>.value(resBody);
      } else  {
        return Future.error("Api status error ${response.statusCode} - ${resBody["status"]}: ${resBody["message"]}");
      }
    } catch (error) {
      return Future.error("Api response error ${error.toString()}");
    }
  }

  Future<List<T>> syncFromRemote<T extends VasatModel>(T clazz, {Map<String, String> params}) async {

    var url = "/api/" + T.toString();
    debugPrint("ENTER syncFromRemote URL: $url");
    var res = apiGet(url, params).then((value) {
      var results = value as List<dynamic>;
      debugPrint("ENETR sync results: " + results.toString());
      debugPrint("------------------------------------------------------");
      return results.map((ele) => clazz.fromJson(ele) as T).toList();
    });

    return res;
  }

  Future<void> saveObjects<T extends VasatModel>(List<T> objects) async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    var objs = objects.map((ele) {
      var obj = ele as VasatModel<T>;
      return jsonEncode( obj.toJson() );
    } ).toList();
    prefs.setStringList(T.toString(), objs);
  }

  Future<List<T>> getObjects<T extends VasatModel<T>>(T classType) async {

    return SharedPreferences.getInstance().then((prefs) {
      var list = prefs.getStringList(T.toString());
      List<T> objList;
      if (list != null) {
        objList = list.map((ele) {
          return classType.fromJson(jsonDecode(ele));
        } ).toList();
      }
      return objList;

    });
  }
}
